#!/usr/bin/python
# Author: Luca Marturana (luca89) <lucamarturana@gmail.com>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# TODO: Improve performance:
#       reduce calls that rewrites things (fill_file_pad, refresh_file_pad)
#       maybe nice to add localizations with gettext

import os
import sys
import re
from datetime import date
import curses
import curses.ascii
from portage import settings as port_settings
from glob import glob

# This text is used on the in-line help
helptext = """
Elogv is a portage elog viewer based on curses and python,
you can use these keys to control the behavior of the program:

 - Down arrow -> Scroll the list of files down by 1 unit
 - Up arrow -> opposite of Down arrow
 - Pag Down -> Scroll the list down by 10 unit
 - Pag Up -> opposite of Pag Down
 - End -> Go to the last file of the list
 - Home -> Go to the first file of the list
 - t -> order the list of files by date, most recent on top
 - a -> order the list of files alphabetically
 - r -> reverse the list of files
 - SpaceBar -> scroll the selected file
 - h or F1 -> show the help screen, press Page Up/Down to scroll up and down,
              h and F1 again to hide
 - d -> removes selected files, usage is similar to vim "d" command,
        here are same examples:
            da -> removes all files
            de -> removes from selected item to the end of the list
            ds -> remove from selected item to the start of the list
            d1d or dd -> removes selected file only
            d4d -> removes 4 files starting from selected one
 - / -> starts a search prompt, write a string and will be showed the next
        package that contains your string, use ESC to exit
 - q -> quit
"""

(normal,selected,einfo,ewarn,eerror,elog) = range(6)

# Exceptions classes
class TermTooSmall(Exception):
    def __init__(self):
        pass

class NoLogFiles(Exception):
    def __init__(self):
        pass

class CannotOpenElogdir(Exception):
    def __init__(self):
        pass

# Main class (called with curses.wrapper later)
class ElogViewer:
    def __init__(self, screen):
        curses.curs_set(0)
        self.screen = screen
        
        # Our color pairs
        curses.init_pair(selected, curses.COLOR_BLACK, curses.COLOR_WHITE)
        curses.init_pair(einfo, curses.COLOR_GREEN, curses.COLOR_BLACK)
        curses.init_pair(ewarn, curses.COLOR_YELLOW, curses.COLOR_BLACK)
        curses.init_pair(eerror, curses.COLOR_RED, curses.COLOR_BLACK)
        curses.init_pair(elog, curses.COLOR_MAGENTA, curses.COLOR_BLACK)
        
        # This attributes are used to manage the scrolling of the list
        # of files    
        self.pposy = 0
        self.usel = 0
        
        # Method used to order the list: by date o by alphabetic order
        # second value set if the list should be reversed (True o False)
        self.sort_method = ["date", False]
        
        # Refer to the file used in self.show_log
        self.f = None
        
        # Initialize screen
        self.init_screen()
        self.show_log()
        
        while True:
            c = self.screen.getch()

            if c == ord('q') or c == curses.ascii.ESC:
                sys.exit()
            
            ## Scrolling keys ##
            elif c == curses.KEY_DOWN:
                self.change_usel(1)
                
            elif c == curses.KEY_UP:
                self.change_usel(-1)

            elif c == curses.KEY_NPAGE:
                self.change_usel(10)
                
            elif c == curses.KEY_PPAGE:
                self.change_usel(-10)

            elif c == curses.KEY_END:
                self.change_usel(len(self.file_list)-1, False)

            elif c == curses.KEY_HOME:
                self.change_usel(0, False)
            ## End Scrolling keys ##
            
            elif c == ord("a"):
                self.sort_method[0] = "alphabetical"
                
            elif c == ord("t"):
                self.sort_method[0] = "date"
                
            elif c == ord("r"):
                self.sort_method[1] = not self.sort_method[1]
                
            elif c == ord(" "):
                # Now is used only for scrolling the text
                self.show_log()

            elif c == curses.KEY_RESIZE:
                # Reinitialize screen
                self.init_screen()
            
            elif c == curses.KEY_F1 or c == ord("h"):
                self.show_help()
                # We need to reinitialize the screen
                self.init_screen()

            elif c == ord("d"):
                n = ''
                while True:
                    subc = self.screen.getch()
                    if subc == ord("a"):
                        n = "all"
                    elif subc == ord("e"):
                        n = "end"
                    elif subc == ord("s"):
                        n = "start"
                    elif curses.ascii.isdigit(subc):
                        n += chr(subc)
                        continue
                    elif subc == ord("d"):
                        if not n:
                            n = "1"
                    break
                
                if n:
                    self.remove_file(n)
                    self.file_pad.clear()
                    self.fill_file_pad()
                    self.refresh_file_pad()

            elif c == ord("/"):
                word = ""
                self.screen.move(self.height-1,1)
                self.screen.addstr("/")
                while True:
                    subc = self.screen.getch()
                    if subc == ord("\n"):
                        self.search(word,1)
                    elif subc == curses.ascii.BS:
                        word = word[:-1]
                        self.screen.delch()
                    elif subc == curses.ascii.ESC:
                        self.screen.hline(self.height-1,1,curses.ACS_HLINE,(len(word)+1))
                        self.screen.addstr(self.height-2,2," "*10)
                        break
                    elif curses.ascii.isalpha(subc):
                        word += chr(subc)
                        self.screen.addstr(chr(subc))
                        self.search(word)  
                
    def init_screen(self):
        """
        Init the screen and wins, it's also used to reinizialize screen
        after a terminal resizing
        """
        (self.height, self.width) = self.screen.getmaxyx()
        
        # Check if the terminal window is too small
        if self.height < 12 or self.width < 80:
            raise TermTooSmall()
            
        # Screen Look&Feel
        self.screen.border()
        self.screen.hline(self.height/2,1, "_", self.width-2)
        m = " Press F1 or h to show the help screen "
        self.screen.addstr(self.height-1, self.width-len(m)-1, m)
        self.screen.refresh()
        
        # Initialize log file window
        self.log_win = curses.newwin(self.height/2-2, self.width-2,
                                     self.height/2+1, 1)
        
        # Draw other window of the screen
        self.fill_file_pad()
        self.refresh_file_pad()
            
    def change_usel(self,n,relative=True):
        prev_usel = self.usel
        if not relative:
            self.usel = n
        elif n < 0 and self.usel+n < 0:
            self.usel = 0
        elif n > 0 and self.usel+n > len(self.file_list)-1:
            self.usel = len(self.file_list)-1
        else:
            self.usel += n
            
        self.file_pad.addstr(prev_usel,0,
                                 "%s - %s" % (self.packages[prev_usel],self.dates[prev_usel]),
                                 curses.color_pair(normal))
        self.file_pad.addstr(self.usel,0,
                                 "%s - %s" % (self.packages[self.usel],self.dates[self.usel]),
                                 curses.color_pair(selected))
        
        first = self.pposy
        last = (self.height / 2 - 2) + first
            
        if self.usel < first:
            self.pposy -= first - self.usel
               
        if self.usel > last:
            self.pposy += self.usel - last
        self.refresh_file_pad()
        self.show_log()
        
    def refresh_file_pad(self):
        """
        Redraws file pad, first half of the screen.
        Can be used to scroll or simply update
        """
        self.file_pad.refresh(self.pposy,0,1,1,self.height/2-1,self.width-2)
        
    def fill_file_pad(self):
        """
        Fill the list of files, colorize the selected row and order files by
        method specified
        """
        # Get the list of files
        try:
            os.chdir(elogdir)
            self.file_list = glob("*:*:*.log")
        except OSError:
            raise CannotOpenElogdir()
        
        if not self.file_list:
            raise NoLogFiles()
        
        # Maybe that after removing files self.usel points to a wrong index,
        # so this will prevent a crash
        if self.usel >= len(self.file_list):
            self.usel = len(self.file_list)-1

        # Sort the list
        if self.sort_method[0] == "alphabetical":
            self.file_list.sort()
        elif self.sort_method[0] == "date":
            def sort_by_date(a,b):
                date_str1 = a.split(":")[2][:8]
                date_str2 = b.split(":")[2][:8]
                
                date1 = date(int(date_str1[:4]),int(date_str1[4:6]),int(date_str1[6:]))
                date2 = date(int(date_str2[:4]),int(date_str2[4:6]),int(date_str2[6:]))
                
                if date1 < date2:
                    retval = 1
                elif date1 == date2:
                    retval = 0
                else:
                    retval = -1
                
                return retval
            
            self.file_list.sort(cmp=sort_by_date)
        
        if self.sort_method[1]:
            self.file_list.reverse()
            
        self.packages = []
        self.dates = []
        self.file_pad = curses.newpad(len(self.file_list),self.width)
        self.file_pad.clear()
        
        for item in self.file_list:
            (category,package_name,other) = item.split(":")
            self.packages.append(category + "/" + package_name)
            self.dates.append(other[:8])
        
        for i in range(0,len(self.file_list)):
            if i == self.usel:
                cp = selected
            else:
                cp = normal
            
            self.file_pad.addstr(i,0,
                                 "%s - %s" % (self.packages[i],self.dates[i]),
                                 curses.color_pair(cp))
        
    def show_log(self):
        """
        Display the selected file, if the length of the file is bigger than
        the height of the window, interrupt the drawing and resume it when the
        user press again the SpaceBar key
        """
        self.log_win.clear()
        
        filepath = os.path.join(elogdir,self.file_list[self.usel])
        if not self.f or self.lastlog != filepath:
            self.lastlog = filepath
            self.f = file(filepath)
        
        for i in range(0,self.height/2-4):
            try:
                x = self.f.next()
            except StopIteration:
                self.f.close()
                self.f = None
                break

            if x.startswith('INFO:'):
                self.log_win.addstr(x[:self.width-2],curses.color_pair(einfo))
            elif x.startswith('WARN:'):
                self.log_win.addstr(x[:self.width-2],curses.color_pair(ewarn))
            elif x.startswith('ERROR:'):
                self.log_win.addstr(x[:self.width-2],curses.color_pair(eerror))
            elif x.startswith('LOG:'):
                self.log_win.addstr(x[:self.width-2],curses.color_pair(elog))
            else:
                self.log_win.addstr(x[:self.width-2],curses.color_pair(normal))
        
        if self.f: 
            self.log_win.addstr(self.height/2-3, self.width-15, "Continue...",
                                curses.color_pair(normal))
            
        self.log_win.refresh()
    
    def remove_file(self,n):
        """
        Delete from the filesystem a slice of elog files
        n can be:
        "all" -> all files will be deleted
        "end" -> files from selected one to the end of the list will be deleted
        <int> -> will be deleted the selected file and the next <int>-1
        """
       
        if n == "all":
            start = None
            end = None
        elif n == "end":
            start = self.usel
            end = None
        elif n == "start":
            start = None
            end = self.usel
        else:
            start = self.usel
            end = self.usel + int(n)
        
        for f in self.file_list[start:end]:
            os.remove(os.path.join(elogdir,f))
    
    def show_help(self):
        # Setup help window
        helpwin_height = self.height / 3 * 2
        helpwin_corner = (self.height / 6, self.width / 2 - 40)
        helpwin = curses.newwin(helpwin_height, 80,
                                     helpwin_corner[0], helpwin_corner[1])
        
        # Setup help pad
        row = 0
        maxrow = len(helptext.splitlines())
        helppad = curses.newpad(maxrow,80)
        
        # Insert helptext on the pad
        # FIXME: variable "i" is ugly
        i = 0
        for x in helptext.splitlines():
            helppad.addstr(i,0,x)
            i += 1
        del i
        
        # Loop to manage user actions
        while True:
            helpwin.clear()
            helpwin.border()
            helpwin.refresh()
            
            helppad.refresh(row,0,helpwin_corner[0]+1,helpwin_corner[1]+1,
                             helpwin_height+helpwin_corner[0]-2,80+helpwin_corner[1]-2)
            c = self.screen.getch()
            
            if c == ord("h") or c == curses.KEY_F1 or c == ord("q"):
                return
            elif c == curses.KEY_NPAGE:
                if row+10 <= maxrow:
                    row += 10
            elif c == curses.KEY_PPAGE:
                if row-10 >= 0:
                    row -= 10
                    
    def search(self, word, div=0):
        for x in self.packages[self.usel+div:]:
            if re.search(word, x):
                self.change_usel(self.packages.index(x),False)
                break
        else:
            self.screen.addstr(self.height-2,2,"Not Found!",
            	curses.color_pair(eerror))
                
if __name__ == "__main__":
    if "--help" in sys.argv:
        print helptext
        sys.exit()
        
    # Get the path of the elogdir
    if port_settings['PORT_LOGDIR']:
        elogdir = os.path.join(port_settings['PORT_LOGDIR'],"elog")
    else:
        elogdir = os.path.join(os.sep,"var","log","portage","elog")
   
    # Launch curses interface
    try:
        curses.wrapper(ElogViewer)
    except TermTooSmall:
        print "Your terminal window is too small, try to enlarge it"
        sys.exit(1)
    except NoLogFiles:
        print "There aren't any elog files on",elogdir
        sys.exit(1)
    except CannotOpenElogdir:
        print "Cannot open",elogdir
        print "Please check if the directory exists and if it's readable by your user."
        sys.exit(1)
    except KeyboardInterrupt:
        pass
